from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

from web3.contract import (
    BaseContractFunctions,
    ContractFunction,
    call_contract_function
)
from web3.types import (
    ABI,
    ABIFunction,
    FunctionIdentifier,
    CallOverride
)

from cfx_utils.token_unit import (
    to_int_if_drip_units
)
from cfx_address import (
    Base32Address
)
from conflux_web3.types import (
    TxParam,
    AddressParam,
    EpochNumberParam,
)
from conflux_web3._utils.contracts import (
    prepare_transaction,
)
from conflux_web3._utils.transactions import (
    fill_transaction_defaults,
)
from conflux_web3._utils.normalizers import (
    addresses_to_verbose_base32
)

if TYPE_CHECKING:
    from conflux_web3 import Web3
    from conflux_web3.types.transaction_hash import TransactionHash

def build_transaction_for_function(
        address: Base32Address,
        web3: 'Web3',
        function_name: Optional[FunctionIdentifier] = None,
        transaction: Optional[TxParam] = None,
        contract_abi: Optional[ABI] = None,
        fn_abi: Optional[ABIFunction] = None,
        *args: Any,
        **kwargs: Any) -> TxParam:
    """Builds a dictionary with the fields required to make the given transaction

    Don't call this directly, instead use :meth:`Contract.build_transaction`
    on your contract instance.
    """
    prepared_transaction:TxParam = prepare_transaction(
        address, # type: ignore
        web3,
        fn_identifier=function_name, # type: ignore
        contract_abi=contract_abi,
        fn_abi=fn_abi,
        transaction=transaction,  # type: ignore
        fn_args=args,
        fn_kwargs=kwargs,
    ) 

    prepared_transaction = fill_transaction_defaults(web3, prepared_transaction)

    return prepared_transaction  # type: ignore

class ConfluxContractFunction(ContractFunction):
    w3: "Web3"
    address: Base32Address
    
    def __call__(self, *args: Any, **kwargs: Any) -> "ConfluxContractFunction":
        return super().__call__(*args, **kwargs) # type: ignore
    
    def build_transaction(self, transaction: Optional[TxParam] = None) -> TxParam:
        built_transaction = self._build_transaction(transaction)  # type: ignore
        return build_transaction_for_function(
            self.address,
            self.w3,
            self.function_identifier,
            built_transaction,  # type: ignore
            self.contract_abi,
            self.abi,
            *self.args,
            **self.kwargs,
        )
    
    def call(self,
            transaction: Optional[TxParam] = None,
            block_identifier: Optional[EpochNumberParam] = "latest_state",
            state_override: Optional[CallOverride] = None,
            ccip_read_enabled: Optional[bool] = None) -> Any:
        call_transaction = self._get_call_txparams(transaction) # type: ignore

        # block_id = parse_block_identifier(self.w3, block_identifier)

        return call_contract_function(
            self.w3,
            self.address, # type: ignore
            # self._return_data_normalizers,
            [
                addresses_to_verbose_base32(self.w3.cfx.chain_id), # type: ignore
            ],
            self.function_identifier,
            call_transaction,
            block_identifier, # type: ignore
            self.contract_abi,
            self.abi,
            state_override,
            ccip_read_enabled,
            *self.args,
            **self.kwargs,
        )

    def transact(self, transaction: Optional[TxParam] = None) -> "TransactionHash":
        if transaction and "value" in transaction:
            transaction["value"] = to_int_if_drip_units(transaction["value"])
        return super().transact(transaction) # type: ignore


class ConfluxContractFunctions(BaseContractFunctions):
    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: Optional[AddressParam] = None,
    ) -> None:
        super().__init__(abi, w3, ConfluxContractFunction, address)  # type: ignore
    
    def __getattr__(self, function_name: str) -> "ConfluxContractFunction":
        return super().__getattr__(function_name) # type: ignore



# Note that ERC1319 is not a finalized standard. We only provide VERY BASIC support for it
from typing import (
    TYPE_CHECKING,
)
from cfxpm.contract import ConfluxLinkableContract

from ethpm.package import Package as EthPackage

if TYPE_CHECKING:
    from conflux_web3 import Web3

# used to hook Package
def set_conflux_linkable_contract(self: "Package", *args, **kwargs):
    self.w3.cfx.defaultContractFactory = ConfluxLinkableContract

class Package(EthPackage):
    w3: "Web3"
        
    def get_contract_factory(self, name: str) -> ConfluxLinkableContract:
        return super().get_contract_factory(name) # type: ignore



from typing import (
    Generic,
    Optional,
    Sequence,
    Callable,
)
from web3.method import (
    Method,
    Munger,
    TFunc
)
from web3.types import (
    RPCEndpoint,
    TReturn
)
from conflux_web3._utils.method_formatters import (
    cfx_request_formatters,
    cfx_result_formatters,
)


class ConfluxMethod(Method[TFunc]):
    def __init__(
        self,
        json_rpc_method: Optional[RPCEndpoint] = None,
        mungers: Optional[Sequence[Munger]] = None,
        request_formatters: Optional[Callable[..., TReturn]] = None,
        result_formatters: Optional[Callable[..., TReturn]] = None,
        null_result_formatters: Optional[Callable[..., TReturn]] = None,
        method_choice_depends_on_args: Optional[Callable[..., RPCEndpoint]] = None,
        is_property: bool = False,
    ):
        request_formatters = request_formatters or cfx_request_formatters  # type: ignore
        result_formatters = result_formatters or cfx_result_formatters # type: ignore
        Method.__init__(self, 
                        json_rpc_method, 
                        mungers, 
                        request_formatters, 
                        result_formatters, 
                        null_result_formatters,
                        method_choice_depends_on_args,
                        is_property
                        )


from typing import (
    Callable,
    Union
)

from web3.module import Module
from cfx_address import Base32Address

from conflux_web3.method import ConfluxMethod
from conflux_web3._utils.rpc_abi import RPC

class Txpool(Module):
    _next_nonce: ConfluxMethod[Callable[[Union[Base32Address, str]], int]] = ConfluxMethod(
        RPC.txpool_nextNonce
    )
    
    def next_nonce(self, address: Union[Base32Address, str]) -> int:
        return self._next_nonce(address)



import json
import os

TAG = "latest"
TESTNET_TAG = "2.1.0-testnet"
REPO_NAME = "confluxchain/conflux-rust"
DEV_IMAGE_FULL_NAME = f"{REPO_NAME}:{TAG}"
TESTNET_IMAGE_FULL_NAME = f"{REPO_NAME}:{TESTNET_TAG}"
LOCAL_NODE_NAME = "python-sdk-env"
TESTNET_NODE_NAME = "python-sdk-env-testnet"
LOCAL_HOST = "127.0.0.1"
PORT = "12537"
TESTNET_HOST_PORT = "12637"

TESTNET_CONFIG_DIR = os.path.join(
    os.path.dirname(__file__), 
    "testnet"
)
VOLUMES = {
    TESTNET_CONFIG_DIR: {
        "bind": "/root/run",
        "mode": "rw",
    }
}

HELPER_DIR = os.path.dirname(__file__)



from typing import Iterable, Sequence, Union
import os
import pytest
from tests._test_helpers.node import (
    LocalNode, BaseNode, RemoteTestnetNode, LocalTestnetNode
)

from cfx_account.account import LocalAccount
from cfx_account import Account
from conflux_web3 import (
    Web3
)
from conflux_web3.types import (
    Base32Address
)
from conflux_web3.middleware import (
    Wallet
)

@pytest.fixture(scope="session")
def use_testnet() -> bool:
    return bool(os.environ.get("TESTNET_SECRET")) or bool(os.environ.get("USE_TESTNET"))

@pytest.fixture(scope="session")
def node(use_testnet) -> Iterable[BaseNode]:
    if use_testnet:
        node = RemoteTestnetNode() # connection error might occur if using public RPC
        # node = LocalTestnetNode() 
        yield node
        # node.exit()
    else:
        node = LocalNode()
        yield node
        # node.exit()

@pytest.fixture(scope="session")
def node_url(node):
    return node.url

@pytest.fixture(scope="session")
def secret_key(node: LocalNode) -> Union[str, None]:
    """
    Returns:
        str: secret key with enough balance
    """
    return node.secrets[0]

# no scope here
@pytest.fixture
def w3(node_url: str, node: LocalNode) -> Web3:
    """
    Returns:
        Web3: a web3 instance
    """
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    return w3

@pytest.fixture(scope="session")
def account(node_url: str, secret_key) -> LocalAccount:
    """external_account, not supported by node
    """
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    acct = w3.account.from_key(secret_key)
    if w3.cfx.get_balance(acct.address) != 0:
        return w3.account.from_key(secret_key)
    elif w3.cfx.chain_id == 1:
        w3.cfx.default_account = acct
        faucet = w3.cfx.contract(name="Faucet")
        faucet.functions.claimCfx().transact().executed()
        return acct
    else:
        raise Exception("Unexpected exception: not local nor testnet node and no environment secret key is set")

@pytest.fixture(scope="module")
def moduled_w3(node_url: str, node: LocalNode, account) -> Web3:
    """
    a web3 instance for the convenience to create module shared objects
    e.g. a transaction or a contract which is required for the module
    NOTE: DON'T CHANGE PROPERTY OF THIS W3
    """
    # account = Account.from_key(secret_key, )
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    w3.cfx.default_account = account
    return w3

@pytest.fixture(scope="session")
def address(node_url, secret_key) -> str:
    chain_id = Web3(Web3.HTTPProvider(node_url)).cfx.chain_id
    addr = Account.from_key(secret_key, chain_id).address
    return addr

@pytest.fixture
def embedded_accounts(w3: Web3, use_testnet: bool) -> Sequence[Base32Address]:
    if use_testnet:
        return []
    return w3.cfx.accounts



from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
)

from eth_typing import (
    TypeStr,
)

from eth_utils.toolz import (
    curry, # type: ignore
)

from cfx_address import (
    Base32Address
)
from conflux_web3._utils.cns import (
    is_cns_name,
    resolve_if_cns_name,
)
from conflux_web3.exceptions import (
    NoWeb3Exception
)

if TYPE_CHECKING:
    from conflux_web3 import Web3  # noqa: F401


@curry
def abi_cns_resolver(w3: "Web3", type_str: TypeStr, val: Any) -> Tuple[TypeStr, Any]:
    if type_str == "address" and is_cns_name(val):
        if w3 is None:
            raise NoWeb3Exception(
                f"Could not look up name {val!r} because no web3"
                " connection available"
            )
        return type_str, resolve_if_cns_name(w3, val)
    else:
        return type_str, val

@curry
def addresses_to_verbose_base32(
    network_id: int, type_str: TypeStr, data: Any
) -> Tuple[TypeStr, Base32Address]:
    if type_str == "address":
        return type_str, Base32Address(data, network_id, verbose=True, _ignore_invalid_type=True)
    return type_str, data



from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Optional,
)
from eth_utils.toolz import (
    assoc, # type: ignore
)
from eth_utils import (
    to_bytes, # type: ignore
)
from ethpm.contract import (
    LinkableContract
)
from ethpm.exceptions import (
    BytecodeLinkingError,
)
from ethpm.contract import (
    is_prelinked_bytecode,
    
)
from conflux_web3.contract import (
    ConfluxContract,
    ConfluxContractConstructor,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

# totally borrows the implementation of LinkableContract except for base class
# used to hack ethpm Package
# hack activated in _web3_hook

# the implementation below does not work because super() won't work
# ConfluxLinkableContract = type(
#     'LinkableContract', (ConfluxContract, ), dict(LinkableContract.__dict__)
# )
class ConfluxLinkableContract(LinkableContract, ConfluxContract):
    def __init__(self, address: bytes, **kwargs: Any) -> None:
        if self.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Contract cannot be instantiated until its bytecode is linked."
            )

        # type ignored to allow for undefined **kwargs on `Contract` base class __init__
        super().__init__(address=address, **kwargs)  # type: ignore

    @classmethod
    def factory(cls, w3: "Web3", class_name: Optional[str] = None, **kwargs: Any) -> ConfluxContract:
        dep_link_refs = kwargs.get("unlinked_references")
        bytecode = kwargs.get("bytecode")
        needs_bytecode_linking = False
        if dep_link_refs and bytecode:
            if not is_prelinked_bytecode(to_bytes(hexstr=bytecode), dep_link_refs):
                needs_bytecode_linking = True
        kwargs = assoc(kwargs, "needs_bytecode_linking", needs_bytecode_linking)
        return super().factory(w3, class_name, **kwargs) # type: ignore

    @classmethod
    def constructor(cls, *args: Any, **kwargs: Any) -> ConfluxContractConstructor:
        if cls.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Contract cannot be deployed until its bytecode is linked."
            )
        return super().constructor(*args, **kwargs) # type: ignore



from typing import (
    TYPE_CHECKING,
    Optional,
)

from web3.contract import (
    BaseContractCaller,
)
from web3.types import (
    ABI,
    CallOverride
)

from conflux_web3.types import (
    TxParam,
    AddressParam,
    EpochNumberParam,
)

from .function import (
    ConfluxContractFunction
)

if TYPE_CHECKING:
    from conflux_web3 import Web3
    

class ConfluxContractCaller(BaseContractCaller):
    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: AddressParam,
        transaction: Optional[TxParam] = None,
        block_identifier: EpochNumberParam = "latest_state",
        ccip_read_enabled: Optional[bool] = None,
    ) -> None:
        super().__init__(
            abi=abi,
            w3=w3,
            address=address, # type: ignore
            transaction=transaction, # type: ignore
            block_identifier=block_identifier, # type: ignore
            ccip_read_enabled=ccip_read_enabled,
            contract_function_class=ConfluxContractFunction,
        )

    def __call__(
        self,
        transaction: Optional[TxParam] = None,
        block_identifier: EpochNumberParam = "latest_state",
        state_override: Optional[CallOverride] = None,
        ccip_read_enabled: Optional[bool] = None,
    ) -> "ConfluxContractCaller":
        if transaction is None:
            transaction = {}
        return type(self)(
            self.abi,
            self.w3, # type: ignore
            self.address,
            transaction=transaction,
            block_identifier=block_identifier,
            ccip_read_enabled=ccip_read_enabled,
        )




from inspect import Parameter
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    cast,
)
from eth_utils.toolz import (
    # assoc,
    curry,  # type: ignore
    merge,  # type: ignore
)

from cfx_utils.token_unit import (
    Drip
)
from conflux_web3.exceptions import (
    NoWeb3Exception
)
from conflux_web3.types import (
    TxParam,
)
from conflux_web3._utils.cns import (
    resolve_if_cns_name,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3


TRANSACTION_DEFAULTS = {
    "value": 0,
    "data": b"",
    "nonce": lambda w3, tx, estimate=None: w3.cfx.get_next_nonce(tx['from']),
    "gas": lambda w3, tx, estimate=None: estimate["gasLimit"],
    "storageLimit": lambda w3, tx, estimate=None: estimate["storageCollateralized"],
    # convert to int value
    "gasPrice": lambda w3, tx, estimate=None: w3.cfx.gas_price.to(Drip).value,
    "chainId": lambda w3, tx, estimate=None: w3.cfx.chain_id,
    "epochHeight": lambda w3, tx, estimate=None: w3.cfx.epoch_number,
}


@curry
def fill_transaction_defaults(w3: "Web3", transaction: TxParam) -> TxParam:
    """
    Fill the necessary fields to "send" a transaction
    Before this function is invoked, ensure 'from' field is filled
    """
    if not w3:
        raise NoWeb3Exception("A web3 object is required to fill transaction defaults, but no web3 object is passed")
    if (not transaction.get("from")) and (transaction.get("nonce", None) is None):
        raise ValueError("Transaction's 'from' field is required to fill nonce field")
    if "from" in transaction:
        transaction['from'] = resolve_if_cns_name(w3, transaction['from'])
    
    for key, default_getter in TRANSACTION_DEFAULTS.items():
        estimate = None
        if key not in transaction:
            if callable(default_getter):
                if not estimate:
                    if key == "gas" or key == "storageLimit":
                        estimate = w3.cfx.estimate_gas_and_collateral(transaction)
                default_val = default_getter(w3, transaction, estimate)
                
            else:
                default_val = default_getter

            transaction.setdefault(key, default_val) # type: ignore
    return transaction



from abc import ABC
import json
import os
from typing import (
    List
)
import time
import urllib3
import functools

import docker
from docker.errors import (
    ImageNotFound,
    NotFound
)

from cfx_account import Account
from conflux_web3 import (
    Web3,
)
from conflux_web3._utils.decorators import (
    cached_property,
)
from tests._test_helpers.ENV_SETTING import (
    DEV_IMAGE_FULL_NAME,
    TESTNET_IMAGE_FULL_NAME,
    LOCAL_NODE_NAME,
    TESTNET_NODE_NAME,
    LOCAL_HOST,
    PORT,
    VOLUMES,
    TESTNET_HOST_PORT
)

def setup_docker_env(client: docker.client.DockerClient, image_name:str, node_name: str):
    try:
        client.images.get(image_name)
    except ImageNotFound:
        client.images.pull(image_name)
    
    # remove the container if existence
    try:
        container = client.containers.get(node_name)
        container.remove(force=True)  # type: ignore
        time.sleep(10)
    except NotFound:
        pass # do nothing
    
def pull_image(client: docker.client.DockerClient, image_name:str):
    try:
        client.images.get(image_name)
    except ImageNotFound:
        client.images.pull(image_name)
    
def get_existed_container(client: docker.client.DockerClient, node_name: str):

    # remove the container if existence
    try:
        container = client.containers.get(node_name)
        # container.remove(force=True)  # type: ignore
        return container
    except NotFound:
        return None
     
def connect_to_server(url):
    # http = urllib3.PoolManager()
    payload = {
        "method": "cfx_getStatus",
        "params": [],
        "jsonrpc": "2.0",
        "id": 1,
    }
    header = {
        "Content-Type": "application/json"
    }
    data = json.dumps(payload)
    http = urllib3.PoolManager()
    res = http.request('POST', url, body=data, headers=header, retries=False)
    result = json.loads(res.data)["result"]
    
    return result

class BaseNode(ABC):
    def __init__(self):
        self._url = None
    
    @property
    def url(self):
        return self._url

    def secrets(self) -> List[str]:
        return []
    
    def exit(self):
        pass

class LocalNode(BaseNode):
    """ using docker to start a private local node.
    if container with node_name (default as "python-sdk-env") already exists, no extra work needs be done
    else pull image and create environment
    """
    def __init__(self, image_name=DEV_IMAGE_FULL_NAME, node_name=LOCAL_NODE_NAME):
        self._image_name = image_name
        self._node_name = node_name
        self._url = f"http://{LOCAL_HOST}:{PORT}"
        self._client = docker.from_env()
        container = get_existed_container(self._client, self._node_name)
        
        if container:
            self._container = container
        else:
            pull_image(self._client, self._image_name)
            self._container = self._client.containers.run(self._image_name, 
                                                        name=self._node_name, 
                                                        detach=True, 
                                                        # auto_remove=True,
                                                        ports={
                                                            f"{PORT}/tcp": f"{PORT}"
                                                        })
            self._wait_for_start()

    @cached_property
    def secrets(self) -> List[str]:
        raw = self._container.exec_run("cat genesis_secret.txt").output.decode("utf-8")  # type: ignore
        secrets = raw.split("\n")
        return secrets
    
    def _wait_for_start(self):
        interval = 1
        max_try = 120
        print("starting test node")
        # time.sleep(10)
        
        for i in range(max_try):
            try:
                status = connect_to_server(self.url)
                if int(status["epochNumber"], 16) >= 1:
                    return self._wait_for_embedded_tx_finished()
                time.sleep(interval)
            except:
                time.sleep(interval)
            # finally:
            #     time.sleep(interval)
        raise ConnectionError("Not connected to node")
    
    def _wait_for_embedded_tx_finished(self):
        interval = 1
        max_try = 300
        # time.sleep(10)
        w3 = Web3(Web3.HTTPProvider(self.url))
        account = w3.account.from_key(self.secrets[0], w3.cfx.chain_id)
        for i in range(max_try):
            if w3.cfx.get_next_nonce(account.address) >= 10:
                return
            time.sleep(interval)
        raise RuntimeError("Exceeds max wait time")
    
    def exit(self):
        if self._container:
            self._container.remove(force=True)  # type: ignore
        if self._client:
            self._client.close()

class LocalTestnetNode(LocalNode):
    def __init__(self, image_name=TESTNET_IMAGE_FULL_NAME, node_name=TESTNET_NODE_NAME, volumes=VOLUMES):
        self._image_name = image_name
        self._node_name = node_name
        self._url = f"http://{LOCAL_HOST}:{TESTNET_HOST_PORT}"
        self._client = docker.from_env()
        container = get_existed_container(self._client, self._node_name)
        if container:
            self._container = container
        else:
            pull_image(self._client, self._image_name)
            self._container = self._client.containers.run(self._image_name, 
                                                        name=self._node_name, 
                                                        detach=True, 
                                                        # auto_remove=True,
                                                        volumes=volumes,
                                                        ports={
                                                            f"{PORT}/tcp": f"{TESTNET_HOST_PORT}" # use a different port on host
                                                        })
            self._wait_for_start()
    
    @cached_property
    def secrets(self) -> List[str]:
        secrets = [Account.create().key.hex()]
        return secrets 
    
    def _wait_for_embedded_tx_finished(self):
        return

class RemoteTestnetNode(BaseNode):
    def __init__(self) -> None:
        self._url = os.environ.get("TESTNET_URL", None) or 'https://test.confluxrpc.com'
    
    @cached_property
    def secrets(self) -> List[str]:
        testnet_secret = os.environ.get("TESTNET_SECRET", None)
        if not testnet_secret:
            return [Account.create().key.hex()]
        return [testnet_secret]



from typing import (
    TYPE_CHECKING, Optional, Type, Union, cast
)
from hexbytes import HexBytes
from web3._utils.method_formatters import (
    to_hexbytes
)

from conflux_web3.exceptions import (
    NoWeb3Exception
)

if TYPE_CHECKING:
    from conflux_web3 import Web3
    from conflux_web3.types import (
        TxReceipt,
        TxData,
    )

def requires_web3(func):
    def inner(self, *args, **kwargs):
        if self._w3 is None:
            raise NoWeb3Exception("No web3 instance is attached to transaction hash."
                        "Use transactionHash.set_w3() to attach a w3 instance")
        return func(self, *args, **kwargs)
    return inner

class TransactionHash(HexBytes):
    _w3: "Web3" = None # type: ignore
    
    def __new__(cls: Type[bytes], val: Union[bool, bytearray, bytes, int, str]) -> "TransactionHash":
        val = to_hexbytes(32, val)
        return cast(TransactionHash, super().__new__(cls, val)) # type: ignore
    
    def set_w3(self, w3: "Web3"):
        self._w3 = w3
    
    # @property
    # def status(self):
    #     pass
    
    # TODO
    # def wait_till(self, target_status: str):
    #     pass
    
    @requires_web3
    def mined(self, timeout: float = 60, poll_latency: float = 0.5) -> "TxData":
        return self._w3.cfx.wait_till_transaction_mined(self, timeout, poll_latency)
    
    @requires_web3
    def executed(self, timeout: float = 300, poll_latency: float = 0.5) -> "TxReceipt":
        return self._w3.cfx.wait_for_transaction_receipt(self, timeout, poll_latency) 
    
    @requires_web3
    def confirmed(self, timeout: float = 600, poll_latency: float = 0.5) -> "TxReceipt":
        return self._w3.cfx.wait_till_transaction_confirmed(self, timeout, poll_latency) 
    
    @requires_web3
    def finalized(self, timeout: float = 1200, poll_latency: float = 1) -> "TxReceipt":
        return self._w3.cfx.wait_till_transaction_finalized(self, timeout, poll_latency)
    
    def __repr__(self) -> str:
        return f"TransactionHash({self.hex()!r})"
    


from typing import (
    TYPE_CHECKING,
    Optional,
    Sequence,
    Tuple,
    cast,
)

from ens.utils import (
    default,
)
from ens.constants import (
    ACCEPTABLE_STALE_HOURS,
)

from cfx_address import (
    Base32Address
)

if TYPE_CHECKING:
    from conflux_web3 import Web3 as _Web3  # noqa: F401
    from web3.providers.base import (  # noqa: F401
        # AsyncBaseProvider,
        BaseProvider,
    )
    from conflux_web3.types import (
        Middleware
    )

def init_web3(
    provider: "BaseProvider" = cast("BaseProvider", default),
    middlewares: Optional[Sequence[Tuple["Middleware", str]]] = None,
    default_account: Optional[Base32Address] = None
) -> "_Web3":
    from conflux_web3 import Web3 as Web3Main
    from conflux_web3.client import ConfluxClient as CfxMain

    if provider is default:
        w3 = Web3Main(ens=None, modules={"cfx": (CfxMain)})   # type: ignore
    else:
        w3 = Web3Main(provider, middlewares, ens=None, modules={"cfx": (CfxMain)})  # type: ignore
    if default_account:
        w3.cfx._default_account = default_account
    return customize_web3(w3)


def customize_web3(w3: "_Web3") -> "_Web3":
    from web3.middleware.stalecheck import make_stalecheck_middleware

    if w3.middleware_onion.get("name_to_address"):
        w3.middleware_onion.remove("name_to_address")

    if not w3.middleware_onion.get("stalecheck"):
        w3.middleware_onion.add(
            make_stalecheck_middleware(ACCEPTABLE_STALE_HOURS * 3600, ("cfx_getBlockByEpochNumber",)), name="stalecheck"
        )
    return w3

# is_none_or_base_32_zero_address is implemented in _web3_hook
# def is_none_or_base_32_zero_address(addr) -> bool:
#     ...


from typing import (
    TYPE_CHECKING, 
    Sequence, 
    Tuple,
)
from conflux_web3.middleware.pending import (
    PendingTransactionMiddleware
)
from conflux_web3.middleware.wallet import (
    Wallet,
    construct_sign_and_send_raw_middleware
)
from conflux_web3.middleware.cache import (
    simple_cache_middleware
)
from conflux_web3.middleware.names import (
    name_to_address_middleware
)
from conflux_web3.types import (
    Middleware
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

def conflux_default_middlewares(w3: "Web3") -> Sequence[Tuple[Middleware, str]]:
    return [
        (name_to_address_middleware(w3), "name_to_address"),
        (PendingTransactionMiddleware, "PendingTransactionMiddleware"),
        (Wallet(), "wallet"),
        (simple_cache_middleware, "CacheMiddleware"),
    ]


__all__ = [
    "PendingTransactionMiddleware",
    "Wallet",
    "construct_sign_and_send_raw_middleware",
    "conflux_default_middlewares"
]



class NoWeb3Exception(Exception):
    pass

class DisabledException(Exception):
    pass

class DeploymentInfoNotFound(Exception):
    pass

class ContractMetadataNotFound(Exception):
    pass

class NameServiceNotSet(Exception):
    pass



import json
import os
from pathlib import (
    Path
)
from typing import (
    Dict,
    Optional
)
from toolz import (
    keyfilter
)

from ens import (
    abis,
)
from cfx_address.utils import (
    validate_address_agaist_network_id
)
from conflux_web3.exceptions import (
    DeploymentInfoNotFound,
    ContractMetadataNotFound
)

METADATA_DIR = Path(__file__).parent

# DEPLOYMENT_INFO
DEPLOYMENT_INFO = {
    "AdminControl": "0x0888000000000000000000000000000000000000",
    "SponsorWhitelistControl": "0x0888000000000000000000000000000000000001",
    "Staking": "0x0888000000000000000000000000000000000002",
    # "ConfluxContext": "0x0888000000000000000000000000000000000003",
    "PoSRegister": "0x0888000000000000000000000000000000000005",
    "CrossSpaceCall": "0x0888000000000000000000000000000000000006",
    "ParamsControl": "0x0888000000000000000000000000000000000007",
    "Create2Factory": "0x8A3A92281Df6497105513B18543fd3B60c778E40",
    "ERC1820": "0x88887eD889e776bCBe2f0f9932EcFaBcDfCd1820",
    "Faucet": "cfxtest:acejjfa80vj06j2jgtz9pngkv423fhkuxj786kjr61",
    "cUSDT": {
        1: "cfxtest:acepe88unk7fvs18436178up33hb4zkuf62a9dk1gv",
        1029: "cfx:acf2rcsh8payyxpg6xj7b0ztswwh81ute60tsw35j7",
    },
    "FC": {
        1: "cfxtest:achkx35n7vngfxgrm7akemk3ftzy47t61yk5nn270s",
        1029: "cfx:achc8nxj7r451c223m18w2dwjnmhkd6rxawrvkvsy2",
    },
    "ENS": {
        1: "cfxtest:acen57mpbzvs774tk6kffcsbkef3m4mn5eh0nxy4jx",
    }
}

METADATA_INFO = {
    "cUSDT": "ERC20",
    "FC": "ERC20",
}

def list_embedded_contract_names():
    pass

# TODO: normalize metadata["bin"] to metadata["bytecode"]
# TODO: return type as TypedDict
def get_contract_metadata(
    contract_name: str, 
    chain_id: Optional[int]=None, 
    with_deployment_info: Optional[bool]=None
) -> Dict:
    """
    _summary_

    Parameters
    ----------
    contract_name : str
        the name of the contract
    chain_id : Optional[int], optional
        _description_, by default None
    with_deployment_info : Optional[bool], optional
        whether "address" will be in the returned Dict.
        if is None, this api will try to fill the address field if possible, and no exception will be raised
        if is False, this api will never fill the address field.
        if is True, this api will always try to fill the address field, and may raise exceptions
        by default None

    Returns
    -------
    Dict
        _description_

    Raises
    ------
    ContractMetadataNotFound
        _description_
    DeploymentInfoNotFound
        _description_
    """
    try:
        metadata_path = METADATA_DIR / f"{METADATA_INFO.get(contract_name, contract_name)}.json"
        if not os.path.exists(metadata_path):
            raise ContractMetadataNotFound(f"Metadata for {contract_name} not found")
        with open(metadata_path) as f:
            metadata = json.load(f)
        metadata = keyfilter(lambda x: x in ["abi", "bytecode"], metadata)
    except ContractMetadataNotFound as e:
        abi = getattr(abis, contract_name, None)
        if abi:
            metadata = {
                "abi": abi
            }
        else:
            raise e
    # process address field if with_deployment_info is True or None
    if with_deployment_info is not False:
        if contract_name not in DEPLOYMENT_INFO:
            if with_deployment_info is True:
                raise DeploymentInfoNotFound(f"Deployment info for {contract_name} not found")
            return metadata
        address_info = DEPLOYMENT_INFO[contract_name]
        # "AdminControl": "0x0888000000000000000000000000000000000000"
        if isinstance(address_info, str):
            validate_address_agaist_network_id(address_info, chain_id, accept_hex=True)
            metadata["address"] = address_info
        # "cUSDT": {
        #     1: "cfxtest:acepe88unk7fvs18436178up33hb4zkuf62a9dk1gv",
        #     1029: "cfx:acf2rcsh8payyxpg6xj7b0ztswwh81ute60tsw35j7",
        # }
        elif isinstance(address_info, dict):
            if not chain_id:
                raise ValueError(f"{contract_name}'s deployment info varies with different chain id, chain id should be specified")
            if with_deployment_info and (chain_id not in address_info):
                raise DeploymentInfoNotFound(f"Deployment info for {contract_name} of chain id {chain_id} not found")
            metadata["address"] = address_info[chain_id]
        else:
            raise Exception("unexpected error")
    return metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    find_packages,
    setup,
)

VERSION = "1.0.1"
DESCRIPTION = 'Python SDK for Conflux network'
with open('./README.md') as readme:
    long_description = readme.read()

extras_require = {
    'tester': [
        "docker>=5.0.0,<6",
        "pytest>=6.2.5,<7",
        "typing_extensions",
        "pytest-cov",
        "ipfshttpclient==0.8.0a2",
        # "py-geth>=3.8.0,<4",
    ],
    'linter': [
        "black>=22.1.0,<23.0",
        # "flake8==3.8.3",
        # "isort>=4.2.15,<4.3.5",
        # "mypy==0.910",
        # "types-setuptools>=57.4.4,<58",
        # "types-requests>=2.26.1,<3",
        # "types-protobuf==3.19.13",
    ],
    'docs': [
        # "mock",
        # "sphinx-better-theme>=0.1.4",
        # "click>=5.1",
        # "configparser==3.5.0",
        # "contextlib2>=0.5.4",
        # "py-geth>=3.8.0,<4",
        # "py-solc>=0.4.0",
        # "pytest>=6.2.5,<7",
        # "sphinx>=4.2.0,<5",
        "jupyter-book==0.11.3",
        # "sphinx_rtd_theme>=0.1.9",
        # "toposort>=1.4",
        # "towncrier==18.5.0",
        # "urllib3",
        "wheel"
    ],
    'dev': [
        # "bumpversion",
        # "flaky>=3.7.0,<4",
        # "hypothesis>=3.31.2,<6",
        # "pytest>=6.2.5,<7",
        # "pytest-asyncio>=0.18.1,<0.19",
        # "pytest-mock>=1.10,<2",
        # "pytest-pythonpath>=0.3",
        # "pytest-watch>=4.2,<5",
        # "pytest-xdist>=1.29,<2",
        # "setuptools>=38.6.0",
        # "tox>=1.8.0",
        # "tqdm>4.32,<5",
        # "twine>=1.13,<2",
        # "pluggy==0.13.1",
        # "when-changed>=0.3.0,<0.4"
    ],
    "ipfs": [
        "ipfshttpclient==0.8.0a2",
    ],
}

extras_require['dev'] = (
    extras_require['tester']
    + extras_require['linter']
    + extras_require['docs']
    # + extras_require["ipfs"] ipfs is included in tester
    + extras_require['dev']
)


# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="conflux-web3",
    version=VERSION,
    author="Conflux-Dev",
    author_email="wenda.zhang@confluxnetwork.org",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    package_data={"conflux_web3": ["contract/metadata/*.json", "py.typed"],
                  "cns": ["py.typed"],
                  "cfxpm": ["py.typed"]},
    url='https://github.com/conflux-chain/python-conflux-sdk',
    install_requires=[
        "web3==6.0.0b9",
        "cfx-address>=1.0.0",
        "cfx-account>=1.0.0",
        "cfx-utils>=1.0.0",
        # "cached_property==1.5.2", # required by cfx-account
        # "eth-account>=0.6.0,<0.7.0"
    ],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'
    extras_require=extras_require,
    keywords=['python', 'conflux', 'blockchain'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from conflux_web3.main import Web3

def get_local_web3():
    return Web3(Web3.HTTPProvider("http://127.0.0.1:12537"))

def get_testnet_web3():
    return Web3(Web3.HTTPProvider("https://test.confluxrpc.com"))

def get_mainnet_web3():
    return Web3(Web3.HTTPProvider("https://main.confluxrpc.com"))
